@@ -33,10 +33,12 @@ gint     gtk_gs_defaults_get_paper_count(void);
 gboolean gtk_gs_defaults_get_show_scroll_rect(void);
 GtkGSPaperSize *gtk_gs_defaults_get_paper_sizes(void);
 const gchar *gtk_gs_defaults_get_interpreter_cmd(void);
+const gchar *gtk_gs_defaults_get_alpha_parameters(void);
 const gchar *gtk_gs_defaults_get_scan_pdf_cmd(void);
 const gchar *gtk_gs_defaults_get_ungzip_cmd(void);
 const gchar *gtk_gs_defaults_get_unbzip2_cmd(void);
 void gtk_gs_defaults_set_interpreter_cmd(gchar *);
+void gtk_gs_defaults_set_alpha_parameters(gchar *);
 void gtk_gs_defaults_set_scan_pdf_cmd(gchar *);
 void gtk_gs_defaults_set_ungzip_cmd(gchar *);
 void gtk_gs_defaults_set_unbzip2_cmd(gchar *);
