@@ -170,6 +170,7 @@ static void gtk_gs_emit_error_msg (GtkGS *gs, const gchar *msg);
 static void gtk_gs_defaults_changed (GConfClient* client, guint cnxn_id,
                                      GConfEntry *entry, gpointer user_data);
 static void send_ps(GtkGS *gs, long begin, unsigned int len, gboolean close);
+static void gtk_gs_set_adjustments(GtkGS *gs, GtkAdjustment *hadj, GtkAdjustment *vadj);
 static void set_up_page (GtkGS *gs);
 static void close_pipe (int p[2]);
 static void interpreter_failed (GtkGS *gs);
@@ -298,6 +299,13 @@ gtk_gs_class_init (GtkGSClass *klass)
         widget_class->unrealize     = gtk_gs_unrealize;
         widget_class->size_request  = gtk_gs_size_request;
         widget_class->size_allocate = gtk_gs_size_allocate;
+        widget_class->set_scroll_adjustments_signal =
+                gtk_signal_new ("set_scroll_adjustments",
+                                GTK_RUN_LAST,
+                                object_class->type,
+                                GTK_SIGNAL_OFFSET(GtkGSClass, set_scroll_adjustments),
+                                gtk_marshal_NONE__POINTER_POINTER,
+                                GTK_TYPE_NONE, 2, GTK_TYPE_ADJUSTMENT, GTK_TYPE_ADJUSTMENT);
 
         /* Create atoms */
         klass->gs_atom        = gdk_atom_intern ("GHOSTVIEW", FALSE);
@@ -309,7 +317,7 @@ gtk_gs_class_init (GtkGSClass *klass)
 
         /* a default handler for "interpreter_message" signal */
         klass->interpreter_message = gtk_gs_interpreter_message;
-
+        klass->set_scroll_adjustments = gtk_gs_set_adjustments;
         gtk_gs_defaults_load();
 }
 
@@ -424,6 +432,11 @@ gtk_gs_destroy (GtkObject *object)
                 gs->gconf_notify_id = -1;
         }
 
+        if(gs->hadj)
+                gtk_object_unref (GTK_OBJECT (gs->hadj));
+        if(gs->vadj)
+                gtk_object_unref (GTK_OBJECT (gs->vadj));
+
         if (GTK_OBJECT_CLASS (parent_class)->destroy)
                 (* GTK_OBJECT_CLASS (parent_class)->destroy) (object);
 }
@@ -439,51 +452,51 @@ gtk_gs_munge_adjustments (GtkGS *gs)
 
         gdk_window_get_position (gs->pstarget, &x, &y);
 
-	/* g_print("Munge adj: waw=%d gsw=%d hau=%d hal=%d \n", gs->widget.allocation.width, gs->width, gs->hadj->upper, gs->hadj->lower);
-	 */
-
 	/* 
 	 * This is a bit messy:
 	 * we want to make sure that we do the right thing if dragged.
 	 */
 	if (gs->widget.allocation.width >= gs->width) {
 		x = (gs->widget.allocation.width - gs->width)/2;
-		gs->hadj->value = (gs->hadj->upper - gs->hadj->lower)/2;
-		gs->hadj->page_size = gs->hadj->upper - gs->hadj->lower;
-	} else {
+		gs->hadj->value = 0.0;
+		gs->hadj->page_size = 1.0;
+	}
+        else {
 		if (x > 0)
 			x = 0;
 		else if (gs->widget.allocation.width > x + gs->width) 
 			x = gs->widget.allocation.width - gs->width;
 		gs->hadj->page_size = ((gfloat) gs->widget.allocation.width)/gs->width;
-		gs->hadj->value = (gs->hadj->upper-gs->hadj->lower) *
-			((gfloat) (gs->widget.allocation.width/2 - x))/gs->width;
+		gs->hadj->value = -((gfloat)x)/gs->width;
 	}
 	if (gs->widget.allocation.height >= gs->height) {
 		y = (gs->widget.allocation.height - gs->height)/2;
-		gs->vadj->value = (gs->vadj->upper - gs->vadj->lower)/2;
-		gs->vadj->page_size = gs->vadj->upper - gs->vadj->lower;
-	} else {
+		gs->vadj->value = 0.0;
+		gs->vadj->page_size = 1.0;
+	}
+        else {
 		if (y > 0)
 			y = 0;
 		else if (gs->widget.allocation.height > y + gs->height) 
 			y = gs->widget.allocation.height - gs->height;
 		gs->vadj->page_size = ((gfloat) gs->widget.allocation.height)/gs->height;
-		gs->vadj->value = (gs->vadj->upper-gs->vadj->lower) *
-			((gfloat) (gs->widget.allocation.height/2 - y))/gs->height;
+		gs->vadj->value = -((gfloat)y)/gs->height;
 	}
 
+        if(gs->scroll_start_x != -1 && gs->scroll_start_y != -1) {
+                gdk_draw_rectangle(gs->pstarget, gs->psgc, FALSE,
+                                   gs->scroll_start_x, gs->scroll_start_y,
+                                   gs->scroll_width, gs->scroll_height);
+        }
+	gdk_window_move(gs->pstarget, x, y);
+        while(gtk_events_pending()) {
+                gtk_main_iteration();
+        }
+        if(gs->scroll_start_x != -1 && gs->scroll_start_y != -1)
+                gdk_draw_rectangle(gs->pstarget, gs->psgc, FALSE,
+                                   gs->scroll_start_x, gs->scroll_start_y,
+                                   gs->scroll_width, gs->scroll_height);
 
-        if(gs->scroll_start_x != -1 && gs->scroll_start_y != -1)
-                gdk_draw_rectangle(gs->pstarget, gs->psgc, FALSE,
-                                   gs->scroll_start_x, gs->scroll_start_y,
-                                   gs->scroll_width, gs->scroll_height);
-	/* now we want to update the slider. */
-	gdk_window_move (gs->pstarget, x, y);
-        if(gs->scroll_start_x != -1 && gs->scroll_start_y != -1)
-                gdk_draw_rectangle(gs->pstarget, gs->psgc, FALSE,
-                                   gs->scroll_start_x, gs->scroll_start_y,
-                                   gs->scroll_width, gs->scroll_height);
 	gtk_adjustment_changed (gs->hadj);
 	gtk_adjustment_changed (gs->vadj);
 }
@@ -610,7 +623,7 @@ gtk_gs_size_allocate (GtkWidget     *widget,
         /* 
          * update the adjustment if necessary (ie. a resize);
          */
-        if ((gs->width > 0) && (gs->height >0) && GTK_WIDGET_REALIZED(gs)) {
+        if ((gs->width > 0) && (gs->height > 0) && GTK_WIDGET_REALIZED(gs)) {
 		gtk_gs_munge_adjustments (gs);
         }
 
@@ -665,22 +678,22 @@ gtk_gs_value_adjustment_changed (GtkAdjustment *adjustment, gpointer data)
    
         gdk_window_get_geometry (gs->pstarget, &x, &y, &width, &height, &depth);
         if (gs->width <= gs->widget.allocation.width)
-                newx = (gs->widget.allocation.width-gs->width)/2;
+                newx = (gs->widget.allocation.width - gs->width)/2;
         else
-                newx = gs->widget.allocation.width/2 -
-                        (gs->hadj->value - gs->hadj->lower) * width /
-                        (gs->hadj->upper - gs->hadj->lower);
+                newx = -gs->hadj->value*gs->width;
         if (gs->height <= gs->widget.allocation.height)
                 newy = (gs->widget.allocation.height-gs->height)/2;
         else
-                newy = gs->widget.allocation.height/2 -
-                        (gs->vadj->value - gs->vadj->lower) * height /
-                        (gs->vadj->upper - gs->vadj->lower);
+                newy = -gs->vadj->value*gs->height;
+
         if(gs->scroll_start_x != -1 && gs->scroll_start_y != -1)
                 gdk_draw_rectangle(gs->pstarget, gs->psgc, FALSE,
                                    gs->scroll_start_x, gs->scroll_start_y,
                                    gs->scroll_width, gs->scroll_height);
-        gdk_window_move (gs->pstarget, newx, newy);
+        gdk_window_move(gs->pstarget, newx, newy);
+        while(gtk_events_pending()) {
+                gtk_main_iteration();
+        }
         if(gs->scroll_start_x != -1 && gs->scroll_start_y != -1)
                 gdk_draw_rectangle(gs->pstarget, gs->psgc, FALSE,
                                    gs->scroll_start_x, gs->scroll_start_y,
@@ -1435,6 +1448,7 @@ compute_size(GtkGS *gs)
         change = (new_width != gs->width * gs->zoom_factor) || (new_height != gs->height * gs->zoom_factor);
         gs->width = (gint) (new_width * gs->zoom_factor);
         gs->height = (gint) (new_height * gs->zoom_factor);
+
         if(GTK_WIDGET_REALIZED(gs)) {
                 if(!gs->loaded) {
                         if(gdk_window_is_visible(gs->pstarget))
@@ -1503,11 +1517,11 @@ gtk_gs_new (GtkAdjustment *hadj, GtkAdjustment *vadj)
         /* we set the adjustment widget to a default size... */
         hadj->lower = 0.0;
         hadj->upper = 1.0;
-        hadj->value = 0.5;
+        hadj->value = 0.0;
         hadj->page_size = 1.0;
         vadj->lower = 0.0;
         vadj->upper = 1.0;
-        vadj->value = 0.5;
+        vadj->value = 0.0;
         vadj->page_size = 1.0;
 
         gtk_signal_connect (GTK_OBJECT (hadj), "value_changed",
@@ -1571,32 +1585,19 @@ gtk_gs_center_page(GtkGS *gs)
 void
 gtk_gs_scroll(GtkGS *gs, gint x_delta, gint y_delta)
 {
-        gfloat hval, vval, uhval, lhval, uvval, lvval;
+        gfloat hval, vval;
 
         g_return_if_fail(gs != NULL);
         g_return_if_fail(GTK_IS_GS(gs));
 
         hval = gs->hadj->value + ((gfloat)x_delta)/gs->width;
         vval = gs->vadj->value + ((gfloat)y_delta)/gs->height;
-        uhval = gs->hadj->upper - gs->hadj->page_size/2;
-        lhval = gs->hadj->lower + gs->hadj->page_size/2;
-        uvval = gs->vadj->upper - gs->vadj->page_size/2;
-        lvval = gs->vadj->lower + gs->vadj->page_size/2;
-        if(hval <= uhval) {
-                if(hval >= lhval)
-                        gtk_adjustment_set_value(gs->hadj, hval);
-                else
-                        gtk_adjustment_set_value(gs->hadj, lhval);
-        } else
-                gtk_adjustment_set_value(gs->hadj, uhval);
-
-        if(vval <= uvval) {
-                if(vval >= lvval)
-                        gtk_adjustment_set_value(gs->vadj, vval);
-                else
-                        gtk_adjustment_set_value(gs->vadj, lvval);
-        } else
-                gtk_adjustment_set_value(gs->vadj, uvval);
+        if(hval <= gs->hadj->upper - gs->hadj->page_size &&
+           hval >= gs->hadj->lower)
+                gtk_adjustment_set_value(gs->hadj, hval);
+        if(vval <= gs->vadj->upper - gs->vadj->page_size &&
+           vval >= gs->vadj->lower)
+                gtk_adjustment_set_value(gs->vadj, vval);
 }
 
 void
@@ -2350,3 +2351,55 @@ gboolean gtk_gs_get_show_scroll_rect(GtkGS *gs)
         return gs->show_scroll_rect;
 }
 
+void
+gtk_gs_set_adjustments (GtkGS         *gs,
+                        GtkAdjustment *hadj,
+                        GtkAdjustment *vadj)
+{
+  g_return_if_fail (gs != NULL);
+  g_return_if_fail (GTK_IS_GS (gs));
+  if (hadj)
+    g_return_if_fail (GTK_IS_ADJUSTMENT (hadj));
+  else
+    hadj = GTK_ADJUSTMENT (gtk_adjustment_new (0.0, 0.0, 1.0, 0.0, 0.0, 1.0));
+  if (vadj)
+    g_return_if_fail (GTK_IS_ADJUSTMENT (vadj));
+  else
+    vadj = GTK_ADJUSTMENT (gtk_adjustment_new (0.0, 0.0, 1.0, 0.0, 0.0, 1.0));
+  
+  if (gs->hadj && (gs->hadj != hadj))
+    {
+      gtk_signal_disconnect_by_data (GTK_OBJECT (gs->hadj), gs);
+      gtk_object_unref (GTK_OBJECT (gs->hadj));
+    }
+  
+  if (gs->vadj && (gs->vadj != vadj))
+    {
+      gtk_signal_disconnect_by_data (GTK_OBJECT (gs->vadj), gs);
+      gtk_object_unref (GTK_OBJECT (gs->vadj));
+    }
+  
+  if (gs->hadj != hadj)
+    {
+      gs->hadj = hadj;
+      gtk_object_ref (GTK_OBJECT (gs->hadj));
+      gtk_object_sink (GTK_OBJECT (gs->hadj));
+      
+      gtk_signal_connect (GTK_OBJECT (hadj), "value_changed",
+                          (GtkSignalFunc) gtk_gs_value_adjustment_changed,
+                          (gpointer) gs);
+
+
+    }
+  
+  if (gs->vadj != vadj)
+    {
+      gs->vadj = vadj;
+      gtk_object_ref (GTK_OBJECT (gs->vadj));
+      gtk_object_sink (GTK_OBJECT (gs->vadj));
+      
+      gtk_signal_connect (GTK_OBJECT (vadj), "value_changed",
+                          (GtkSignalFunc) gtk_gs_value_adjustment_changed,
+                          (gpointer) gs);
+    }
+}
